# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Generic, TypeVar

import numpy as np

BoardGameT = TypeVar("BoardGameT")  # An instance of a board game


class BoardGamePlayer(Generic[BoardGameT]):
    def __init__(self, seed: int | None = None):
        pass

    def play(self, game: BoardGameT) -> int:
        """Return an action to play."""
        raise NotImplementedError()

    def evaluate(self, game: BoardGameT) -> tuple[np.ndarray, dict[int, float]]:
        """An optional method evaluating the game state and returning predicted policy and chosen value estimates.

        Returns:
          (policy, value): The `policy` is a distribution over valid actions; the `value`
            is a dictionary assigning value estimate to chosen actions (possibly empty).
        """
        raise NotImplementedError()



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import gymnasium as gym
import numpy as np


# Discrete environments.
class DiscretizationWrapper(gym.ObservationWrapper):
    def __init__(self, env, separators, tiles=None):
        super().__init__(env)
        self._separators = separators
        self._tiles = tiles

        if tiles is None:
            states = 1
            for separator in separators:
                states *= 1 + len(separator)
            self.observation_space = gym.spaces.Discrete(states)
        else:
            self._first_tile_states, self._rest_tiles_states = 1, 1
            for separator in separators:
                self._first_tile_states *= 1 + len(separator)
                self._rest_tiles_states *= 2 + len(separator)
            self.observation_space = gym.spaces.MultiDiscrete([
                self._first_tile_states + i * self._rest_tiles_states for i in range(tiles)])

            self._separator_offsets = [0 if len(s) <= 1 else (s[1] - s[0]) / tiles for s in separators]
            self._separator_tops = [np.inf if len(s) <= 1 else s[-1] + (s[1] - s[0]) for s in separators]

    def observation(self, observations):
        state = 0
        for observation, separator in zip(observations, self._separators):
            state *= 1 + len(separator)
            state += np.digitize(observation, separator)
        if self._tiles is None:
            return state
        else:
            states = np.empty(self._tiles, dtype=np.int64)
            states[0] = state
            for t in range(1, self._tiles):
                state = 0
                for i in range(len(self._separators)):
                    state *= 2 + len(self._separators[i])
                    value = observations[i] + ((t * (2 * i + 1)) % self._tiles) * self._separator_offsets[i]
                    if value > self._separator_tops[i]:
                        state += 1 + len(self._separators[i])
                    else:
                        state += np.digitize(value, self._separators[i])
                states[t] = self._first_tile_states + (t - 1) * self._rest_tiles_states + state
            return states


class DiscreteCartPoleWrapper(DiscretizationWrapper):
    def __init__(self, env, bins=8):
        super().__init__(env, [
            np.linspace(-2.4, 2.4, num=bins + 1)[1:-1],  # cart position
            np.linspace(-3, 3, num=bins + 1)[1:-1],      # cart velocity
            np.linspace(-0.2, 0.2, num=bins + 1)[1:-1],  # pole angle
            np.linspace(-2, 2, num=bins + 1)[1:-1],      # pole angle velocity
        ])


class DiscreteMountainCarWrapper(DiscretizationWrapper):
    def __init__(self, env, bins=None, tiles=None):
        if bins is None:
            bins = 24 if tiles is None or tiles <= 1 else 12 if tiles <= 3 else 8
        super().__init__(env, [
            np.linspace(-1.2, 0.6, num=bins + 1)[1:-1],    # car position
            np.linspace(-0.07, 0.07, num=bins + 1)[1:-1],  # car velocity
        ], tiles)


class DiscreteLunarLanderWrapper(DiscretizationWrapper):
    def __init__(self, env):
        super().__init__(env, [
            np.linspace(-.4, .4, num=5 + 1)[1:-1],      # position x
            np.linspace(-.075, 1.35, num=6 + 1)[1:-1],  # position y
            np.linspace(-.5, .5, num=5 + 1)[1:-1],      # velocity x
            np.linspace(-.8, .8, num=7 + 1)[1:-1],      # velocity y
            np.linspace(-.2, .2, num=3 + 1)[1:-1],      # rotation
            np.linspace(-.2, .2, num=5 + 1)[1:-1],      # ang velocity
            [.5],                                       # left contact
            [.5],                                       # right contact
        ])

        self._expert = gym.make("LunarLander-v3")
        gym.Env.reset(self._expert.unwrapped, seed=42)

    def expert_trajectory(self, seed=None):
        state, trajectory, done = self._expert.reset(seed=seed)[0], [], False
        while not done:
            action = gym.envs.box2d.lunar_lander.heuristic(self._expert, state)
            next_state, reward, terminated, truncated, _ = self._expert.step(action)
            trajectory.append((self.observation(state), action, reward))
            done = terminated or truncated
            state = next_state
        trajectory.append((self.observation(state), None, None))
        return trajectory


# Plotting performance using matplotlib.
class LivePlotWrapper(gym.Wrapper):
    def __init__(self, env, plot_each):
        import matplotlib.pyplot as plt

        super().__init__(env)

        assert plot_each > 0
        self._plot_each = plot_each
        self._return = 0
        self._returns = np.zeros(plot_each)
        self._ret_index = 0

        self._episode_means = []
        self._episode_minus_stds = []
        self._episode_plus_stds = []
        self._mean_ep_indices = []

        self._figure, self._axis = plt.subplots()
        self._figure.show()

    def _update_plot(self):
        self._axis.cla()
        self._axis.fill_between(
            self._mean_ep_indices,
            self._episode_minus_stds,
            self._episode_plus_stds,
            alpha=0.2
        )
        self._axis.plot(self._mean_ep_indices, self._episode_means)
        self._axis.set_xlabel("Episode")
        self._axis.set_ylabel("Return")
        self._axis.grid(True)
        self._figure.canvas.draw()
        self._figure.canvas.flush_events()

    def _add_return(self):
        self._returns[self._ret_index] = self._return
        self._ret_index += 1
        self._return = 0

        if self._ret_index == self._plot_each:
            mean = self._returns.mean()
            std = self._returns.std()

            self._mean_ep_indices.append(len(self._episode_means))
            self._episode_means.append(mean)
            self._episode_minus_stds.append(mean - std)
            self._episode_plus_stds.append(mean + std)

            self._ret_index = 0
            self._update_plot()

    def reset(self, *args, **kwargs):
        self._return = 0
        return self.env.reset(*args, **kwargs)

    def step(self, action):
        transition = self.env.step(action)
        _, reward, terminated, truncated, _ = transition
        self._return += float(reward)
        if terminated or truncated:
            self._add_return()
        return transition

    def save_figure(self, fname, *args, **kwargs):
        """See `plt.savefig` for arguments description."""
        self._figure.savefig(fname, *args, **kwargs)


# Register custom environments.
gym.envs.register(
    id="MountainCar1000-v0",
    entry_point="gymnasium.envs.classic_control.mountain_car:MountainCarEnv",
    max_episode_steps=1000,
    reward_threshold=-110.0,
)



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pygame

from . import cart_pole_pixels


# Allow running the environment and controlling it with arrows
if __name__ == "__main__":
    env = cart_pole_pixels.CartPolePixels(render_mode="human")
    env.metadata["render_fps"] = 10

    quit = False
    while not quit:
        env.reset()
        steps, action, restart = 0, 0, False
        while True:
            # Handle input
            for event in pygame.event.get():
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_LEFT:
                        action = 0
                    if event.key == pygame.K_RIGHT:
                        action = 1
                    if event.key == pygame.K_RETURN:
                        restart = True
                    if event.key == pygame.K_ESCAPE:
                        quit = True
                if event.type == pygame.QUIT:
                    quit = True

            # Perform the step
            _, _, terminated, truncated, _ = env.step(action)

            steps += 1
            if terminated or truncated or restart or quit:
                break
        print("Episode ended after {} steps".format(steps))

    env.close()



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np
import torch


def typed_torch_function(device, *types, via_np=False):
    """Typed Torch function decorator.

    The positional input arguments are converted to torch Tensors of the given
    types and on the given device; for NumPy arrays on the same device,
    the conversion should not copy the data.

    The torch Tensors generated by the wrapped function are converted back
    to Numpy arrays before returning (while keeping original tuples, lists,
    and dictionaries).
    """
    def check_typed_torch_function(wrapped, args):
        if len(types) != len(args):
            while hasattr(wrapped, "__wrapped__"):
                wrapped = wrapped.__wrapped__
            raise AssertionError("The typed_torch_function decorator for {} expected {} arguments, but got {}".format(
                wrapped, len(types), len(args)))

    def structural_map(value):
        if isinstance(value, torch.Tensor):
            return value.numpy(force=True)
        if isinstance(value, tuple):
            return tuple(structural_map(element) for element in value)
        if isinstance(value, list):
            return [structural_map(element) for element in value]
        if isinstance(value, dict):
            return {key: structural_map(element) for key, element in value.items()}
        return value

    class TypedTorchFunctionWrapper:
        def __init__(self, func):
            self.__wrapped__ = func

        def __call__(self, *args, **kwargs):
            check_typed_torch_function(self.__wrapped__, args)
            return structural_map(self.__wrapped__(
                *[torch.as_tensor(np.asarray(arg) if via_np else arg, dtype=typ, device=device)
                  for arg, typ in zip(args, types)], **kwargs))

        def __get__(self, instance, cls):
            return TypedTorchFunctionWrapper(self.__wrapped__.__get__(instance, cls))

    return TypedTorchFunctionWrapper



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import gymnasium as gym
import numpy as np

from . import multi_collect


class SingleCollect(multi_collect.MultiCollect):
    def __init__(self, render_mode: str | None = None):
        super().__init__(agents=1, render_mode=render_mode)
        self.action_space = gym.spaces.Discrete(self.action_space.nvec[0])

    def step(self, action: list[int] | np.ndarray):
        action = np.asarray(action)
        if not action.shape:
            action = np.expand_dims(action, axis=0)
        return super().step(action)


#################################
# Environment for NPFL139 class #
#################################

gym.envs.register(id="SingleCollect-v0", entry_point=SingleCollect, max_episode_steps=250, reward_threshold=0)



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import gymnasium as gym
import gymnasium.envs.classic_control
import numpy as np

import pygame
import pygame.gfxdraw


class CartPolePixels(gymnasium.envs.classic_control.CartPoleEnv):
    H, W = 80, 80
    SCALE = 8

    metadata = {
        "render_modes": ["human", "rgb_array"],
        "render_fps": 50,
    }

    def __init__(self, render_mode=None):
        super().__init__()

        self.render_mode = render_mode
        self._images = 3
        self._image = np.zeros([self.H, self.W, self._images], dtype=np.uint8)
        self._image_surface = pygame.Surface((self.H, self.W))
        self._screen = None

        self.observation_space = gym.spaces.Box(low=0, high=255, shape=[self.H, self.W, self._images], dtype=np.uint8)

    def reset(self, seed=None, options=None):
        render_mode = self.render_mode
        self.render_mode = None
        state, info = super().reset(seed=seed, options=options)
        self.render_mode = render_mode

        for step in range(self._images):
            observation = self._draw(state)

        if self.render_mode == "human":
            self.render()
        return observation, info

    def step(self, action):
        render_mode = self.render_mode
        self.render_mode = None
        state, reward, termination, truncation, info = super().step(action)
        self.render_mode = render_mode

        observation = self._draw(state)

        if self.render_mode == "human":
            self.render()
        return observation, reward, termination, truncation, info

    def render(self):
        assert self.render_mode in self.metadata["render_modes"]

        if self.render_mode == "rgb_array":
            return np.copy(self._image)

        if self._screen is None:
            pygame.init()
            pygame.display.init()
            self._screen = pygame.display.set_mode((self.W * self.SCALE, self.H * self.SCALE))
            self._screen_surface = pygame.Surface((self.W * self.SCALE, self.H * self.SCALE))
            self._clock = pygame.time.Clock()

        pygame.pixelcopy.array_to_surface(
            self._screen_surface, self._image.repeat(8, axis=0).repeat(8, axis=1).transpose([1, 0, 2]))

        pygame.event.pump()
        self._screen.blit(self._screen_surface, (0, 0))
        pygame.display.flip()
        self._clock.tick(self.metadata["render_fps"])
        pygame.event.pump()

    def close(self):
        if self._screen is not None:
            self._screen = None
            pygame.display.quit()
            pygame.quit()

    def _draw(self, state):
        for i in range(self._images - 1):
            self._image[:, :, i] = self._image[:, :, i + 1]

        cart = int(40 + state[0] / 3 * 40)
        cart_poly = [(70, cart - 10), (80, cart - 10), (80, cart + 10), (70, cart + 10)]

        pole_x = int(40 + (state[0] + np.sin(state[2]) * 4.2) / 3 * 40)
        pole_y = int(70 - np.cos(state[2]) * 5.2 / 3 * 40)
        pole_poly = [(pole_y, pole_x - 2), (70, cart - 2), (70, cart + 2), (pole_y, pole_x + 2)]

        self._image_surface.fill([0, 0, 0])
        pygame.gfxdraw.polygon(self._image_surface, pole_poly, [255, 255, 255])
        pygame.gfxdraw.filled_polygon(self._image_surface, pole_poly, [255, 255, 255])
        pygame.gfxdraw.filled_polygon(self._image_surface, cart_poly, [128, 128, 128])
        pygame.pixelcopy.surface_to_array(self._image[:, :, -1], self._image_surface, kind="R")

        return np.copy(self._image)


#################################
# Environment for NPFL139 class #
#################################

gym.envs.register(
    id="CartPolePixels-v1",
    entry_point=CartPolePixels,
    max_episode_steps=500,
    reward_threshold=475,
)



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
if __name__ == "__main__":
    import argparse

    import ale_py  # noqa: F401
    import gymnasium as gym
    import gymnasium.utils.play

    parser = argparse.ArgumentParser()
    parser.add_argument("atari_game", type=str, help="The name of the Atari game to play.")
    parser.add_argument("--frame_skip", default=1, type=int, help="The frameskip amount.")
    parser.add_argument("--zoom", default=4, type=int, help="The zoom level.")
    args = parser.parse_args()

    class ReturnReporter():
        def __init__(self):
            self.rewards = 0

        def __call__(self, obs_t, obs_tp1, action, rew, terminated, truncated, info):
            self.rewards += rew
            if terminated or truncated:
                print("Episode reward:", self.rewards)
                self.rewards = 0

    gym.utils.play.play(
        gym.make("ALE/{}-v5".format(args.atari_game), render_mode="rgb_array",
                 frameskip=args.frame_skip, full_action_space=True),
        callback=ReturnReporter(), fps=60 / args.frame_skip, zoom=args.zoom)



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .board_game import BoardGame
from .board_game_player import BoardGamePlayer
from .board_game_evaluator import evaluate

from .az_quiz import AZQuiz
from .az_quiz_player_fork_heuristic import AZQuizPlayerForkHeuristic
from .az_quiz_player_keyboard import AZQuizPlayerKeyboard
from .az_quiz_player_mouse import AZQuizPlayerMouse
from .az_quiz_player_random import AZQuizPlayerRandom
from .az_quiz_player_simple_heuristic import AZQuizPlayerSimpleHeuristic



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np
import pygame

from . import car_racing

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--frame_skip", default=1, type=int, help="Consider only each --frame_skip frame")
    parser.add_argument("--state", default=False, action="store_true", help="Show state instead of rendering")
    args = parser.parse_args()

    a = np.array([0.0, 0.0, 0.0])

    def register_input():
        global quit, restart
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_LEFT:
                    a[0] = -0.6
                if event.key == pygame.K_RIGHT:
                    a[0] = +0.6
                if event.key == pygame.K_UP:
                    a[1] = +0.5
                if event.key == pygame.K_DOWN:
                    a[2] = +0.8
                if event.key == pygame.K_RETURN:
                    restart = True
                if event.key == pygame.K_ESCAPE:
                    quit = True

            if event.type == pygame.KEYUP:
                if event.key == pygame.K_LEFT:
                    a[0] = 0
                if event.key == pygame.K_RIGHT:
                    a[0] = 0
                if event.key == pygame.K_UP:
                    a[1] = 0
                if event.key == pygame.K_DOWN:
                    a[2] = 0

            if event.type == pygame.QUIT:
                quit = True

    env = car_racing.CarRacingFS(args.frame_skip, render_mode=None if args.state else "human")

    quit = False
    while not quit:
        s, info = env.reset()
        total_reward = 0.0
        steps = 0
        restart = False
        while True:
            if args.state:
                env.show_state(s)
            register_input()
            s, r, terminated, truncated, info = env.step(a)
            total_reward += r
            if steps % 200 == 0 or terminated or truncated or restart or quit:
                print("\naction " + str([f"{x:+0.2f}" for x in a]))
                print(f"step {steps} total_reward {total_reward:+0.2f}")
            steps += 1
            if terminated or truncated or restart or quit:
                break
    env.close()



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np

from .az_quiz import AZQuiz
from .board_game_player import BoardGamePlayer


# Utils
CENTER = 12
CORNER_STONES = [16, 19]
LINE_3 = [11, 12, 13]
CURVE_3 = [11, 12, 18]
BOTTOM_LINE = [16, 17, 18, 19]

def intersect(test1, test2):
    return list(set(test1).intersection(set(test2)))

def union(test1, test2):
    return list(set(test1).union(set(test2)))

# Rotations
BOARD_ROTATIONS = np.array([
    [ 0,  0, 21, 21, 27, 27],
    [ 1,  2, 22, 15, 20, 26],
    [ 2,  1, 15, 22, 26, 20],
    [ 3,  5, 23, 10, 14, 25],
    [ 4,  4, 16, 16, 19, 19],
    [ 5,  3, 10, 23, 25, 14],
    [ 6,  9, 24,  6,  9, 24],
    [ 7,  8, 17, 11, 13, 18],
    [ 8,  7, 11, 17, 18, 13],
    [ 9,  6,  6, 24, 24,  9],
    [10, 14, 25,  3,  5, 23],
    [11, 13, 18,  7,  8, 17],
    [12, 12, 12, 12, 12, 12],
    [13, 11,  7, 18, 17,  8],
    [14, 10,  3, 25, 23,  5],
    [15, 20, 26,  1,  2, 22],
    [16, 19, 19,  4,  4, 16],
    [17, 18, 13,  8,  7, 11],
    [18, 17,  8, 13, 11,  7],
    [19, 16,  4, 19, 16,  4],
    [20, 15,  1, 26, 22,  2],
    [21, 27, 27,  0,  0, 21],
    [22, 26, 20,  2,  1, 15],
    [23, 25, 14,  5,  3, 10],
    [24, 24,  9,  9,  6,  6],
    [25, 23,  5, 14, 10,  3],
    [26, 22,  2, 20, 15,  1],
    [27, 21,  0, 27, 21,  0],
])

INVERSE_ROTATION = [0, 1, 4, 3, 2, 5]

def rotate_list(l, r):
    s = np.zeros(28)
    s[l] = 1
    rs = rotate_situation(s, r)
    return rs.nonzero()[0]

def rotate_situation(situation, r=0):
    rotation = BOARD_ROTATIONS[:, r]
    rotated_situation = np.zeros(len(situation))
    for i in range(len(situation)):
        rotated_situation[i] = situation[rotation[i]]
    return rotated_situation

def all_variants(func):
    def inner(my, enemy, allowed, generator):
        rotations = INVERSE_ROTATION.copy()
        generator.shuffle(rotations)
        possible_moves = []
        for rotation in rotations:
            rotated_my = rotate_list(my, rotation)
            rotated_enemy = rotate_list(enemy, rotation)
            rotated_allowed = rotate_list(allowed, rotation)
            action = func(rotated_my, rotated_enemy, rotated_allowed, generator)
            if action>-1:
                action = BOARD_ROTATIONS[action, rotation]
            if action>=0:
                possible_moves.append(action)
        if len(possible_moves)>0:
            m = intersect(possible_moves, allowed)
            m = generator.choice(m)
        else:
            m=-1
        return m
    return inner

# Rules
def choose_random(arr, allowed, generator):
    if len(intersect(arr,allowed))==0:
        return -1
    return generator.choice(intersect(arr,allowed))

def vidlicky(enemy, allowed, pole, generator):
    if len(intersect(enemy, pole))>=1:
        return choose_random(pole, allowed, generator)
    return -1

@all_variants
def priprav_vidle(my, enemy, allowed, generator):
    if len(intersect(my, [13,12,7]))==3:
        if len(intersect(enemy, [17,18,19]))==3:
            if len(my)==3 and len(enemy)==3:
                return 10
    return -1

@all_variants
def obchazi_vidle_5(my, enemy, allowed, generator):
    if len(intersect(my, [13,12,7, 10]))==4:
        if len(intersect(enemy, [17,18,19]))==3:
            if 11 in enemy and 6 in allowed: return 6
            if 6 in enemy and 11 in allowed: return 11
        if len(intersect(enemy, BOTTOM_LINE))==4 and 15 in allowed: return 15
    if len(intersect(my, [13,12,7, 10, 15]))==5:
        if len(intersect(enemy, BOTTOM_LINE))==4:
            if 22 in enemy and 21 in allowed: return 21
            if 21 in enemy and 22 in allowed: return 22
    if len(intersect(my,[ 8, 11, 12])) ==3:
         if len(intersect(enemy,[13, 17, 18]))==3:
             if 16 in allowed: return 16
    if len(intersect(my,[ 8, 11, 12, 16])) ==4:
         if len(intersect(enemy,[13, 17, 18, 23]))==4:
             if 22 in allowed: return 22

    return -1

@all_variants
def obsad_prostredek(my, enemy, allowed, generator):
    if len(intersect(my, LINE_3))>=1 and len(intersect(enemy, LINE_3))==0:
        return choose_random(LINE_3, allowed, generator)
    return -1

@all_variants
def prostredni_cara(my, enemy, allowed, generator):
    if len(intersect(my, LINE_3))>=2 and len(intersect(enemy, LINE_3))==0:
        return choose_random(LINE_3, allowed, generator)
    return -1

@all_variants
def prostredni_krivka(my, enemy, allowed, generator):
    if len(intersect(my, [11,12]))==2 and 13 in enemy:
        return choose_random([8,18], allowed, generator)
    return -1

@all_variants
def obchazi_vidle_1(my, enemy, allowed, generator):
    if len(intersect(my, LINE_3))==3:
        return vidlicky(enemy, allowed, [10,6], generator)
    return -1

@all_variants
def obchazi_vidle_2(my, enemy, allowed, generator):
    if len(intersect(my, CURVE_3))==3:
        return vidlicky(enemy, allowed, [24,25], generator)
    return -1

@all_variants
def obchazi_vidle_3(my, enemy, allowed, generator):
    if len(intersect(my, CURVE_3))==3:
        return vidlicky(enemy, allowed, [10,6], generator)
    return -1

@all_variants
def obchazi_vidle_4(my, enemy, allowed, generator):
    if 16 in my:
        return vidlicky(enemy, allowed, [16], generator)
    return -1

@all_variants
def obchazi_vidle_6(my, enemy, allowed, generator):
    if len(intersect(my, [7,8,13,18]))==4:
        ret = -1

        if ret==-1: ret = vidlicky(enemy, allowed, [3,6], generator)
        if ret==-1: ret = vidlicky(enemy, allowed, [5,9], generator)
        if ret==-1: ret = vidlicky(enemy, allowed, [24,25], generator)
        return ret

    return -1

@all_variants
def pozor_dole(my, enemy, allowed, generator):
    if len(intersect(my, LINE_3))==3 and len(intersect(enemy, BOTTOM_LINE))>=3:
        return choose_random(BOTTOM_LINE, allowed, generator)
    return -1

@all_variants
def pribliz_se_ke_stedu(my, enemy, allowed, generator):
    if len(intersect(my, LINE_3))==3 and len(intersect(my, [17,18]))==0:
        if 17 in allowed: return 17
        if 18 in allowed: return 18
    return -1

@all_variants
def spoj_stranu_stredem(my, enemy, allowed, generator):
    if len(intersect(my, BOTTOM_LINE))>0:
        for field in intersect(my, BOTTOM_LINE):
            if field+6 in allowed: return field+6
            if field+7 in allowed: return field+7
    return -1

@all_variants
def spoj_stranu_rohem(my, enemy, allowed, generator):
    if len(intersect(my, [15,22]))>0:
        if 21 in allowed: return 21
    return -1

@all_variants
def corner_stone(my, enemy, allowed, generator):
    if len(intersect(my, CORNER_STONES))==0:
        return choose_random(CORNER_STONES, allowed, generator)
    return -1

@all_variants
def corner_stones(my, enemy, allowed, generator):
    if len(intersect(enemy, [16, 19]))==2:
        if 4 in allowed: return 4
    return -1

@all_variants
def dolni_cara(my, enemy, allowed, generator):
    if len(intersect(my, CORNER_STONES))==2:
        if len(intersect(enemy, BOTTOM_LINE))==0:
            return choose_random(BOTTOM_LINE, allowed, generator)
    return -1

@all_variants
def sikma_cara(my, enemy, allowed, generator):
    if len(intersect(enemy, [18,13]))==0:
        return choose_random([18,13], allowed, generator)
    if len(intersect(enemy, [18,13]))==2:
        if 8 in allowed: return 8
    if len(intersect(enemy, [18,13,8]))==3:
        if 4 in allowed: return 4
        if 7 in allowed: return 7
        if 5 in allowed: return 5
    if len(intersect(enemy, [18,13,8,5]))==4:
        if 2 in allowed: return 2
        if 2 in my and 21 in allowed: return 21
    return -1

def hraj_nahodne(my, enemy, allowed, generator):
    return choose_random(allowed, allowed, generator)

@all_variants
def prodluz_dolni_caru(my, enemy, allowed, generator):
    if len(intersect(my, BOTTOM_LINE))==4:
        if len(intersect(enemy, [14,20]))==1:
            return choose_random([14,20], allowed, generator)
        if len(intersect(enemy, [15,10]))==1:
            return choose_random([15,10], allowed, generator)
    return -1

@all_variants
def utok_bez_duvodu(my, enemy, allowed, generator):
    if len(intersect(my, [7,11,8,17]))==4:
        if len(intersect(allowed, [6,3,5,9,23,24]))==6:
            return choose_random([6,3,5,9,23,24], allowed, generator)
    return -1

@all_variants
def lepsi_pulkolecko(my, enemy, allowed, generator):
    if len(intersect(union(allowed,my), [11,17,7,8]))==4 and len(intersect(union(allowed,enemy), [18,13]))==2:
        return choose_random([11,17,7,8], allowed, generator)
    return -1

chain_mam_stred = [
    utok_bez_duvodu,
    obchazi_vidle_1,
    obchazi_vidle_2,
    obchazi_vidle_3,
    obchazi_vidle_4,
    obchazi_vidle_5,
    priprav_vidle,
    pozor_dole,
    prostredni_krivka,
    pribliz_se_ke_stedu,
    prostredni_cara,
    spoj_stranu_stredem,
    spoj_stranu_rohem,
    obsad_prostredek,
    hraj_nahodne,
]

chain_nemam_stred = [
    obchazi_vidle_6,
    utok_bez_duvodu,
    lepsi_pulkolecko,
    corner_stone,
    corner_stones,
    prodluz_dolni_caru,
    sikma_cara,
    dolni_cara,
    spoj_stranu_stredem,
    spoj_stranu_rohem,
    hraj_nahodne,
]

def apply_rules(rules, my, enemy, allowed, generator):
    action = -1
    for rule in rules:
        if action==-1:
            action = rule(my, enemy, allowed, generator)
    return action


class AZQuizPlayerForkHeuristic(BoardGamePlayer[AZQuiz]):
    """An elaborate heuristic for AZQuiz."""
    def __init__(self, seed: int | None = None):
        self._generator = np.random.RandomState(seed)

    def play(self, game: AZQuiz) -> int:
        board = game.board_features
        situation = np.array([board[y, x] for y in range(7) for x in range(y + 1)])
        my = situation[:, game.to_play].nonzero()[0]
        enemy = situation[:, 1 - game.to_play].nonzero()[0]
        taken = union(my, enemy)
        allowed = [action for action in range(28) if action not in taken]
        if CENTER in allowed:
            return CENTER
        if CENTER in enemy:
            return apply_rules(chain_nemam_stred, my, enemy, allowed, self._generator)
        if CENTER in my:
            return apply_rules(chain_mam_stred, my, enemy, allowed, self._generator)


AZQuiz.register_player("fork_heuristic", AZQuizPlayerForkHeuristic)



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .az_quiz import AZQuiz
from .board_game_player import BoardGamePlayer


class AZQuizPlayerKeyboard(BoardGamePlayer[AZQuiz]):
    """An interactive player using keyboard to select an action."""
    def play(self, game: AZQuiz) -> int:
        return game.keyboard_input()


AZQuiz.register_player("keyboard", AZQuizPlayerKeyboard)



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np

from .board_game import BoardGame
from .board_game_player import BoardGamePlayer


class AZQuiz(BoardGame):
    ACTIONS = 28
    N = 7
    C = 4

    def __init__(self, randomized=False):
        self._board = np.tri(self.N, dtype=np.int8) - 1
        self._randomized = randomized
        self._to_play = 0
        self._winner = None
        self._screen = None
        self._last_action, self._winning_stones = None, None

    def clone(self, swap_players=False) -> "AZQuiz":
        clone = AZQuiz(self._randomized)
        if swap_players:
            clone._board = self._SWAP_PLAYERS[self._board + 1]
            clone._to_play = 1 - self._to_play
            clone._winner = 1 - self._winner if self._winner is not None else None
        else:
            clone._board[:, :] = self._board
            clone._to_play = self._to_play
            clone._winner = self._winner
        clone._last_action, clone._winning_stones = self._last_action, self._winning_stones
        return clone

    @property
    def board(self) -> np.ndarray:
        return self._board

    @property
    def board_features(self) -> np.ndarray:
        return self._REPRESENTATION[self._board + 1]

    @property
    def to_play(self) -> int:
        return self._to_play

    def outcome(self, player: int) -> BoardGame.Outcome | None:
        if self._winner is None:
            return None
        return self.Outcome.WIN if self._winner == player else self.Outcome.LOSS

    def valid(self, action: int) -> bool:
        return self._winner is None and action >= 0 and action < self.ACTIONS \
            and self._board[self._ACTION_Y[action], self._ACTION_X[action]] < 2

    def valid_actions(self) -> list[int]:
        return np.nonzero(self._board[self._ACTION_Y, self._ACTION_X] < 2)[0] if self._winner is None else []

    def move(self, action: int):
        self._last_action = action
        return self._move(action, np.random.uniform() if self._randomized else 0)

    def all_moves(self, action):
        success = self.clone()
        success._move(action, 0.)
        if not self._randomized:
            return [(1.0, success)]

        failure = self.clone()
        failure._move(action, 1.)
        if self._board[self._ACTION_Y[action], self._ACTION_X[action]] == 0:
            success_probability = self._INITIAL_QUESTION_PROB
        else:
            success_probability = self._ADDITIONAL_QUESTION_PROB
        return [(success_probability, success), (1. - success_probability, failure)]

    def _move(self, action, random_value):
        if not self.valid(action):
            raise ValueError("An invalid action to AZQuiz.move")

        if self._board[self._ACTION_Y[action], self._ACTION_X[action]] == 0:
            if random_value <= self._INITIAL_QUESTION_PROB:
                self._board[self._ACTION_Y[action], self._ACTION_X[action]] = 2 + self._to_play
            else:
                self._board[self._ACTION_Y[action], self._ACTION_X[action]] = 1
        else:
            if random_value > self._ADDITIONAL_QUESTION_PROB:
                self._to_play = 1 - self._to_play
            self._board[self._ACTION_Y[action], self._ACTION_X[action]] = 2 + self._to_play
        self._to_play = 1 - self._to_play

        edges, visited = np.zeros(2, dtype=np.uint8), np.zeros([self.N, self.N], dtype=np.uint8)
        for j in range(self.N):
            edges[:] = False
            field = self._board[j, 0]
            if field >= 2:
                self._traverse(j, 0, field, edges, visited)
                if edges.all():
                    self._winner = field - 2
                    self._winning_stones = visited == 1
                visited += visited > 0

    def _traverse(self, j, i, field, edges, visited):
        if visited[j, i]: return
        visited[j, i] = 1

        if j == i: edges[0] = True
        if j == self.N - 1: edges[1] = True
        if j - 1 >= 0:
            if i - 1 >= 0 and self._board[j - 1, i - 1] == field: self._traverse(j - 1, i - 1, field, edges, visited)
            if self._board[j - 1, i] == field: self._traverse(j - 1, i, field, edges, visited)
        if i - 1 >= 0 and self._board[j, i - 1] == field: self._traverse(j, i - 1, field, edges, visited)
        if i + 1 < self.N and self._board[j, i + 1] == field: self._traverse(j, i + 1, field, edges, visited)
        if j + 1 < self.N:
            if self._board[j + 1, i] == field: self._traverse(j + 1, i, field, edges, visited)
            if i + 1 < self.N and self._board[j + 1, i + 1] == field: self._traverse(j + 1, i + 1, field, edges, visited)

    def _render(self):
        A = 40
        I = 35
        W = A * 13
        COLORS = np.array([[210, 210, 210], [58, 58, 58], [58, 147, 192], [254, 147, 17],
                           [158, 158, 158], [77, 195, 255], [255, 194, 77]], dtype=np.uint8)

        image = np.zeros([W, W, 3], dtype=np.uint8)
        for j in range(self.N):
            for i in range(j + 1):
                x = int((i - j/2) * 2 * 0.866 * A + W/2)
                y = int((j - (self.N - 1)/2) * 1.5 * A + W/2)
                for yo in range(-I, I + 1):
                    xo = int(min(2 * 0.866 * (I - abs(yo)), 0.866 * I))
                    image[x - xo:x + xo + 1, y + yo] = COLORS[self._board[j, i] + 3 * (
                        (j * (j + 1) // 2 + i == self._last_action) or
                        (self._winning_stones is not None and self._winning_stones[j, i]))]

        return image

    def render(self):
        import pygame

        image = self._render()

        if self._screen is None:
            pygame.init()
            self._screen = pygame.display.get_surface() or pygame.display.set_mode(image.shape[:2])
            self._screen_surface = pygame.Surface(image.shape[:2])

        pygame.pixelcopy.array_to_surface(self._screen_surface, image)
        self._screen.blit(self._screen_surface, (0, 0))
        pygame.display.flip()
        pygame.event.pump()

    def mouse_input(self):
        import pygame

        centers, A, I, W = [], 40, 35, 40 * 13
        for j in range(self.N):
            for i in range(j + 1):
                centers.append(((i - j/2) * 2 * 0.866 * A + W/2,
                                (j - (self.N - 1)/2) * 1.5 * A + W/2))

        self.render()
        chosen = None
        while chosen is None:
            event = pygame.event.wait()
            if event.type == pygame.MOUSEBUTTONDOWN:
                for action, (x, y) in enumerate(centers):
                    distance2 = (event.pos[0] - x)**2 + (event.pos[1] - y)**2
                    if distance2 < I * I and self.valid(action):
                        chosen = action
                        break
            if event.type == pygame.WINDOWEXPOSED:
                self._screen.blit(self._screen_surface, (0, 0))
                pygame.display.flip()

        return chosen

    def keyboard_input(self):
        SYMBOLS = [".", "*", "O", "X"]

        board, action = [], 0
        for j in range(self.N):
            board.append("")
            for mode in range(2):
                board[-1] += "  " * (self.N - 1 - j)
                for i in range(j + 1):
                    board[-1] += " " + (SYMBOLS[self._board[j, i]] * 2 if mode == 0 else "{:2d}".format(action + i)) + " "
                board[-1] += "  " * (self.N - 1 - j)
            action += j + 1

        print("\n".join(board), flush=True)

        action = None
        while action is None or not self.valid(action):
            try:
                action = int(input("Choose action for player {}: ".format(SYMBOLS[2 + self.to_play])))
            except ValueError:
                pass
        return action

    @staticmethod
    def player_from_name(player_name: str) -> type[BoardGamePlayer["AZQuiz"]]:
        assert player_name in AZQuiz._players, f"AZQuiz.player_from_name got unknown player name: {player_name}"
        return AZQuiz._players[player_name]

    @staticmethod
    def register_player(player_name: str, player: type[BoardGamePlayer["AZQuiz"]]):
        AZQuiz._players[player_name] = player

    _players: dict[str, type[BoardGamePlayer["AZQuiz"]]] = {}

    _INITIAL_QUESTION_PROB = 0.8
    _ADDITIONAL_QUESTION_PROB = 0.7

    _ACTION_Y = np.array([0,1,1,2,2,2,3,3,3,3,4,4,4,4,4,5,5,5,5,5,5,6,6,6,6,6,6,6], dtype=np.int8)
    _ACTION_X = np.array([0,0,1,0,1,2,0,1,2,3,0,1,2,3,4,0,1,2,3,4,5,0,1,2,3,4,5,6], dtype=np.int8)
    _REPRESENTATION = np.array([[0,0,0,0], [0,0,0,1], [0,0,1,1], [1,0,0,1], [0,1,0,1]], dtype=np.uint8)
    _SWAP_PLAYERS = np.array([-1,0,1,3,2])


# Explicit class for the randomized version of the game.
class AZQuizRandomized(AZQuiz):
    def __init__(self):
        super().__init__(randomized=True)


# Register both AZQuiz variants.
BoardGame.register_game("az_quiz", AZQuiz)
BoardGame.register_game("az_quiz_randomized", AZQuizRandomized)



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Generic, Sequence, TypeVar

import numpy as np

NamedTuple = TypeVar("NamedTuple")  # A generic type fulfilling the NamedTuple protocol.


class MonolithicReplayBuffer(Generic[NamedTuple]):
    """A monolithic replay buffer with a limited capacity.

    The individual items must be named tuples of data convertible to Numpy
    arrays, each with a fixed shape. The whole replay buffer stores items in an
    efficient manner by keeping a single named tuple of Numpy arrays containing
    all the data.

    Parameters:
      max_length: The maximum number of items the replay buffer can store.
      seed: An optional random seed for sampling; fresh one created if None.
    """
    def __init__(self, max_length: int, seed: int | None = None) -> None:
        assert isinstance(max_length, int), "The max_length argument must be an integer"
        assert max_length > 0, "The max_length argument must be a positive integer"

        self._len: int = 0
        self._max_length: int = max_length
        self._offset: int = 0  # Used when the buffer is full and overwriting in a circular manner.
        self._data: NamedTuple | None = None
        self._generator = np.random.default_rng(seed)

    def __len__(self) -> int:
        """Return the number of items in the replay buffer."""
        return self._len

    @property
    def max_length(self) -> int:
        """Return the maximum capacity of the replay buffer."""
        return self._max_length

    @property
    def data(self) -> NamedTuple | None:
        """Return the data stored in the replay buffer as a named tuple of Numpy arrays."""
        return self._data

    def __getitem__(self, index: int | np.ndarray) -> NamedTuple:
        """Return the item or items at the given index or indices.

        Returns:
          A NamedTuple of Numpy arrays containing the item(s) at the given index(es).
        """
        return self._data._make(value[index] for value in self._data)

    def append(self, item: NamedTuple) -> int:
        """Append a new item to the replay buffer.

        If the buffer is empty, it is allocated to full capacity using Numpy
        arrays of shapes and data types of the provided item.

        Returns:
          The index at which the item was stored.
        """
        # Allocate the buffer on the first append.
        if not self._len:
            self._data = item._make([np.empty((self._max_length, *value.shape), dtype=value.dtype)
                                     for value in map(np.asarray, item)])

        # Select the index to store the new item, updating the sizes.
        if self._len < self._max_length:
            index, self._len = self._len, self._len + 1
        else:
            index, self._offset = self._offset, (self._offset + 1) % self._max_length

        # Store the new item on the given index.
        for i, value in enumerate(item):
            self._data[i][index] = value

        return index

    def append_batch(self, batch: NamedTuple) -> None:
        """Append a named tuple with a batch of items to the replay buffer."""
        if not self._len:
            self._data = batch._make([np.empty((self._max_length, *value.shape[1:]), dtype=value.dtype)
                                      for value in map(np.asarray, batch)])

        batch_size = len(batch[0])
        assert all(len(value) == batch_size for value in batch), "The batch must contain items of the same size"
        assert batch_size > 0, "The batch must contain at least one item"

        if self._len + batch_size <= self._max_length:
            # Append the batch to the end of the buffer in a single call.
            for i, value in enumerate(batch):
                self._data[i][self._len:self._len + batch_size] = value
            self._len += batch_size
        elif self._len == self._max_length and self._offset + batch_size <= self._max_length:
            # Overwrite the oldest items in the buffer in a single call.
            for i, value in enumerate(batch):
                self._data[i][self._offset:self._offset + batch_size] = value
            self._offset = (self._offset + batch_size) % self._max_length
        else:
            # Add batch items one by one to handle wrap-around.
            for batch_index in range(batch_size):
                # Select the index to store the new item, updating the sizes.
                if self._len < self._max_length:
                    index, self._len = self._len, self._len + 1
                else:
                    index, self._offset = self._offset, (self._offset + 1) % self._max_length
                # Store the new item on the given index.
                for i, value in enumerate(batch):
                    self._data[i][index] = value[batch_index]

    def extend(self, items: Sequence[NamedTuple]) -> None:
        """Append a sequence of items to the replay buffer.

        Equivalent to calling append() for each item in the sequence."""
        for item in items:
            self.append(item)

    def sample(self, size: int, replace: bool = True) -> NamedTuple:
        """Sample a batch of items from the replay buffer without replacement.

        Parameters:
          size: The number of items to sample.
          replace: If True, the same item can be sampled multiple times, but the
            random number generator is about twice as fast in this mode.

        Returns:
          A named tuple of Numpy arrays containing the sampled items.
        """
        if replace:
            indices = self._generator.integers(0, self._len, size)
        else:
            indices = self._generator.choice(self._len, size, replace=False)
        return self[indices]



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Generic, Sequence, TypeVar

import numpy as np

Item = TypeVar("Item")  # A generic type for the items in the replay buffer.


class ReplayBuffer(Generic[Item]):
    """Simple replay buffer with possibly limited capacity.

    We use a custom implementation instead of `collections.deque`, which has
    linear complexity of indexing (it is a two-way linked list). The following
    implementation has similar runtime performance as a numpy array of objects,
    but it has unnecessary memory overhead (hundreds of MBs for 1M elements).
    Using five numpy arrays (for state, action, reward, done, and next state)
    would provide minimal memory overhead, but is less flexible.
    """
    def __init__(self, max_length: int | None = None):
        if max_length is not None:
            assert isinstance(max_length, int), "The max_length argument must be an integer"
            assert max_length > 0, "The max_length argument must be a positive integer"
        self._max_length = max_length
        self._data = []
        self._offset = 0

    def __len__(self) -> int:
        return len(self._data)

    @property
    def max_length(self) -> int | None:
        return self._max_length

    def append(self, item: Item) -> None:
        if self._max_length is not None and len(self._data) >= self._max_length:
            self._data[self._offset] = item
            self._offset = (self._offset + 1) % self._max_length
        else:
            self._data.append(item)

    def extend(self, items: Sequence[Item]) -> None:
        if self._max_length is None:
            self._data.extend(items)
        else:
            for item in items:
                if len(self._data) >= self._max_length:
                    self._data[self._offset] = item
                    self._offset = (self._offset + 1) % self._max_length
                else:
                    self._data.append(item)

    def __getitem__(self, index: int) -> Item:
        assert -len(self._data) <= index < len(self._data)
        return self._data[(self._offset + index) % len(self._data)]

    def sample(self, size: int, generator=np.random, replace: bool = True) -> list[Item]:
        # By default, the same element can be sampled multiple times. Making sure the samples
        # are unique is costly, and we do not mind the duplicites much during training.
        if replace:
            return [self._data[index] for index in generator.randint(len(self._data), size=size)]
        else:
            return [self._data[index] for index in generator.choice(len(self._data), size=size, replace=False)]



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import numpy as np

from .az_quiz import AZQuiz
from .board_game_player import BoardGamePlayer


class AZQuizPlayerSimpleHeuristic(BoardGamePlayer[AZQuiz]):
    """A simple heuristic for AZQuiz."""
    CENTER = 12
    ANCHORS = [4, 16, 19]

    def __init__(self, seed: int | None = None):
        super().__init__()
        self._generator = np.random.RandomState(seed)

    def play(self, game: AZQuiz) -> int:
        if game.valid(self.CENTER):
            return self.CENTER

        any_anchor = any(map(game.valid, self.ANCHORS))

        action = None
        while action is None or not game.valid(action):
            if any_anchor:
                action = self._generator.choice(self.ANCHORS)
            else:
                action = self._generator.randint(game.ACTIONS)

        return action


AZQuiz.register_player("simple_heuristic", AZQuizPlayerSimpleHeuristic)



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import gymnasium as gym
import numpy as np
import pygame

if __name__ == "__main__":
    env = gym.make("SingleCollect-v0", render_mode="human")
    env.reset()

    quit = False
    while not quit:
        env.reset()
        steps, action, restart, rewards = 0, 0, False, []
        while True:
            # Handle input
            for event in pygame.event.get():
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_LEFT:
                        action = 2
                    if event.key == pygame.K_RIGHT:
                        action = 1
                    if event.key == pygame.K_UP:
                        action = 3
                    if event.key == pygame.K_DOWN:
                        action = 4
                    if event.key == pygame.K_RETURN:
                        restart = True
                    if event.key == pygame.K_ESCAPE:
                        quit = True
                if event.type == pygame.QUIT:
                    quit = True

            # Perform the step
            _, reward, terminated, truncated, _ = env.step(action)
            if terminated or truncated or restart or quit:
                break

            steps += 1
            action = 0
            rewards.append(reward)
            if len(rewards) % 25 == 0:
                print("Rewards for last 25 timesteps: {}".format(np.sum(rewards[-25:])))
        print("Episode ended with a return of {}".format(np.sum(rewards)))

    env.close()



#!/usr/bin/env python3
#
# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import time

from .board_game import BoardGame
from .board_game_player import BoardGamePlayer


def evaluate(
    game_class: type[BoardGame], players: tuple[BoardGamePlayer, BoardGamePlayer],
    games: int, first_chosen: bool, render: bool = False, verbose: bool = False,
) -> float:
    if first_chosen:
        if game_class.__name__.startswith("AZQuiz"):
            assert games % game_class.ACTIONS == 0, \
                "If `first_chosen` is True, the number of games must be divisble by the number of actions"

            def first_move_selector(game_index: int) -> int:
                return game_index % game_class.ACTIONS
        else:
            raise ValueError(f"The game {game_class.__name__} does not support first move selection")

    wins = [0, 0]
    for i in range(games):
        for to_start in range(2):
            game = game_class()
            if first_chosen:
                game.move(first_move_selector(i))
            while not game.outcome(game.to_play):
                game.move(players[to_start ^ game.to_play].play(game.clone()))
                if render:
                    game.render()
                    time.sleep(0.2)
            if game.outcome(to_start) == game.Outcome.WIN:
                wins[to_start] += 1
            if render:
                time.sleep(2.0)

        if verbose:
            g = i + 1
            print("First player win rate after {} games: {:.2f}% ({:.2f}% and {:.2f}% when starting and not starting)"
                  .format(2 * g, 100 * (wins[0] + wins[1]) / (2 * g), 100 * wins[0] / g, 100 * wins[1] / g))

    return (wins[0] + wins[1]) / (2 * games)



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os
import random

import numpy as np
import torch

def update_params_by_ema(target: torch.nn.Module, source: torch.nn.Module, tau: float) -> None:
    """Update target parameters using exponential moving average of the source parameters.

    Parameters:
      target: The target model whose parameters will be updated.
      source: The source model whose parameters will be used for the update.
      tau: The decay factor for the exponential moving average, e.g., 0.001.
    """
    with torch.no_grad():
        for target_param, source_param in zip(target.parameters(), source.parameters()):
            target_param.mul_(1 - tau)
            target_param.add_(tau * source_param)



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import enum
import numpy as np


class BoardGame:
    class Outcome(enum.Enum):
        LOSS = 1
        DRAW = 2
        WIN = 3

    ACTIONS: int  # Number of possible actions
    N: int  # Size of the board
    C: int  # Number of channels in the board representation

    def clone(self, swap_players=False) -> "BoardGame":
        """Clone the game state, optionally swapping the players."""

    @property
    def board(self) -> np.ndarray:
        """Return the internal representation of board as a NxN numpy array."""

    @property
    def board_features(self) -> np.ndarray:
        """Return the board as a NxNxC numpy array of features."""

    @property
    def to_play(self) -> int:
        """Return the current player."""

    def outcome(self, player: int) -> Outcome | None:
        """Return the game outcome for a given player."""

    def valid(self, action: int) -> bool:
        """Return whether the given action is valid."""

    def valid_actions(self) -> list[int]:
        """Return the list of valid actions."""

    def move(self, action: int) -> None:
        """Execute the given action."""

    # Static factory methods
    @staticmethod
    def from_name(game_name: str) -> type["BoardGame"]:
        """Return a game class from the game name."""
        assert game_name in BoardGame._games, f"BoardGame.from_name got unknown game name: {game_name}"
        return BoardGame._games[game_name]

    @staticmethod
    def register_game(game_name: str, game_class: type["BoardGame"]) -> None:
        """Register a new game class."""
        BoardGame._games[game_name] = game_class

    _games: dict[str, type["BoardGame"]] = {}

    # A game might provide reference players.
    @staticmethod
    def player_from_name(self, player_name: str):
        """Return a player object for the given player name."""



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# EvaluationEnv
from .evaluation_env import EvaluationEnv

# Custom environments
from . import envs

# Environment wrappers
from .env_wrappers import DiscreteCartPoleWrapper
from .env_wrappers import DiscreteLunarLanderWrapper
from .env_wrappers import DiscreteMountainCarWrapper
from .env_wrappers import LivePlotWrapper

# Board games
from . import board_games

# Utils
from .initializers_override import global_keras_initializers
from .monolithic_replay_buffer import MonolithicReplayBuffer
from .replay_buffer import ReplayBuffer
from .startup import startup
from .typed_torch_function import typed_torch_function
from .update_params_by_ema import update_params_by_ema
from .version import __version__, require_version



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import gymnasium as gym
import numpy as np


class MultiCollect(gym.Env):
    metadata = {"render_modes": ["human", "rgb_array"], "render_fps": 25}

    def __init__(self, agents: int, render_mode: str | None = None):
        assert agents > 0
        self._agents = agents

        self.observation_space = gym.spaces.Box(
            np.array([*[-10, -10] * agents, *[-np.inf, -np.inf] * agents, *[-0.5, -0.5] * agents], np.float32),
            np.array([*[+10, +10] * agents, *[+np.inf, +np.inf] * agents, *[+0.5, +0.5] * agents], np.float32))
        self.action_space = gym.spaces.MultiDiscrete([5] * agents)
        self.render_mode = render_mode
        self._screen = None
        self._surface = None

    def reset(self, seed=None, options=None):
        self._agents_pos = self.np_random.uniform(-10, 10, size=[self._agents, 2]).astype(np.float32)
        self._agents_vel = self.np_random.integers(-4, 4 + 1, size=[self._agents, 2])
        self._centers = self.np_random.uniform(-10, 10, size=[self._agents, 2]).astype(np.float32)
        self._centers_hit = np.zeros(self._agents, dtype=np.int32)

        observation, _, _, _, info = self.step(np.zeros(self._agents, dtype=np.int32))

        if self.render_mode == "human":
            self.render()
        return observation, info

    def step(self, action: list[int] | np.ndarray):
        action = np.asarray(action)
        assert len(action) == self._agents
        assert np.all(action >= 0) and np.all(action < 5)

        # Update speeds
        self._agents_vel[:, 0] += (action == 1)
        self._agents_vel[:, 0] -= (action == 2)
        self._agents_vel[:, 1] += (action == 3)
        self._agents_vel[:, 1] -= (action == 4)
        self._agents_vel = np.clip(self._agents_vel, -4, 4)

        # Update pos
        self._agents_pos += self._agents_vel / 8

        # Compute reward and update the hit information
        rewards = np.zeros(self._agents)
        distances = np.linalg.norm(self._centers[:, np.newaxis] - self._agents_pos[np.newaxis, :], axis=-1)
        for i in range(self._agents):
            a = np.argmin(distances[i])
            closest = distances[i][a]
            if closest < 1:
                rewards[a] += 1
                self._centers_hit[i] += 1
                if self._centers_hit[i] >= 10:
                    rewards[a] += 50
                    self._centers_hit[i] = 0
                    self._centers[i] = self.np_random.uniform(-10, 10, size=2)
            else:
                rewards[a] += 1 - (closest - 1) / 10
                self._centers_hit[i] = 0

        distances = np.linalg.norm(self._agents_pos[:, np.newaxis] - self._agents_pos[np.newaxis, :], axis=-1)
        closest = np.min(distances + np.eye(self._agents), axis=1)
        self._agents_hit = closest < 1
        rewards -= self._agents_hit
        observation = np.concatenate(
            [self._centers.ravel(), self._agents_pos.ravel(), self._agents_vel.ravel().astype(np.float32) / 8])

        if self.render_mode == "human":
            self.render()
        return observation, np.mean(rewards), False, False, {"agent_rewards": rewards}

    def render(self):
        return self._render(self.render_mode)

    def _render(self, mode: str):
        import pygame
        import pygame.gfxdraw

        assert mode in self.metadata["render_modes"]

        W, R = 600, 30
        if self._surface is None:
            self._surface = pygame.Surface((W, W))
            if mode == "human":
                pygame.init()
                self._screen = pygame.display.set_mode((W, W))
                self._clock = pygame.time.Clock()

        self._surface.fill((255, 255, 255))
        for (x, y), hit in zip(self._centers, self._centers_hit):
            for method in [pygame.gfxdraw.aacircle, pygame.gfxdraw.filled_circle]:
                method(self._surface, W // 2 + int(x * R), W // 2 - int(y * R), R, (0, 255 if hit else 180, 0))
        for (x, y), hit in zip(self._agents_pos, self._agents_hit):
            for method in [pygame.gfxdraw.aacircle, pygame.gfxdraw.filled_circle]:
                method(self._surface, W // 2 + int(x * R), W // 2 - int(y * R), R // 2,
                       (255, 0, 0) if hit else (0, 0, 180))
        for (x, y), (xv, yv) in zip(self._agents_pos, self._agents_vel):
            pygame.draw.line(self._surface, (0, 0, 0), (W // 2 + int(x * R), W // 2 - int(y * R)),
                             (W // 2 + int((x + xv) * R), W // 2 - int((y + yv) * R)), width=2)

        if mode == "human":
            pygame.event.pump()
            self._clock.tick(self.metadata["render_fps"])
            self._screen.blit(self._surface, (0, 0))
            pygame.display.flip()
        else:
            return np.transpose(
                np.array(pygame.surfarray.pixels3d(self._surface)), axes=(1, 0, 2)
            )

    def close(self):
        if self._screen is not None:
            import pygame
            pygame.display.quit()
            pygame.quit()


#################################
# Environment for NPFL139 class #
#################################

gym.envs.register(id="MultiCollect-v0", entry_point=MultiCollect, max_episode_steps=250, reward_threshold=0)



#!/usr/bin/env python3
#
# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import argparse
import importlib
import os

from .board_game import BoardGame
from .board_game_evaluator import evaluate


def load_player(args: argparse.Namespace, player: str):
    player, *player_args = player.split(":")

    # Try a reference player first
    try:
        return BoardGame.from_name(args.game).player_from_name(player)(args.seed)
    except Exception:
        pass

    # Otherwise, load a player from a module
    if player.endswith(".py"):
        player = player[:-3]

    def loader():
        module = importlib.import_module(player)
        module_args = module.parser.parse_args(player_args)
        module_args.recodex = True
        if hasattr(module_args, "seed") and module_args.seed is None and args.seed is not None:
            module_args.seed = args.seed
        try:
            cwd = os.getcwd()
            os.chdir(os.path.dirname(module.__file__))
            return module.main(module_args)
        finally:
            os.chdir(cwd)

    if args.multiprocessing:
        import multiprocessing

        class Player:
            def __init__(self):
                self._conn, child_conn = multiprocessing.Pipe()
                self._p = multiprocessing.Process(target=Player._worker, args=(child_conn, loader), daemon=True)
                self._p.start()

            def _worker(conn, loader):
                player = loader()
                while True:
                    conn.send(player.play(conn.recv()))

            def play(self, game):
                self._conn.send(game)
                return self._conn.recv()
        return Player()

    else:
        return loader()


if __name__ == "__main__":
    import numpy as np

    parser = argparse.ArgumentParser()
    parser.add_argument("player_1", type=str, help="First player module")
    parser.add_argument("player_2", type=str, help="Second player module")
    parser.add_argument("--first_chosen", default=False, action="store_true", help="The first move is chosen")
    parser.add_argument("--game", default="az_quiz", type=str, help="Game to evaluate")
    parser.add_argument("--games", default=56, type=int, help="Number of alternating games to evaluate")
    parser.add_argument("--multiprocessing", default=False, action="store_true", help="Load players in sep. processes")
    parser.add_argument("--render", default=False, action="store_true", help="Should the games be rendered")
    parser.add_argument("--seed", default=None, type=int, help="Random seed")
    args = parser.parse_args()

    np.random.seed(args.seed)

    evaluate(
        game_class=BoardGame.from_name(args.game),
        players=(load_player(args, args.player_1), load_player(args, args.player_2)),
        games=args.games,
        first_chosen=args.first_chosen,
        render=args.render,
        verbose=True,
    )



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from . import car_racing
from . import cart_pole_pixels
from . import multi_collect
from . import single_collect



# This file is part of NPFL139 <http://github.com/ufal/npfl139/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os
import sys

import gymnasium as gym
import numpy as np


class EvaluationEnv(gym.Wrapper):
    def __init__(self, env, seed=None, render_each=0, evaluate_for=100, report_each=10):
        super().__init__(env)
        self._render_each = render_each
        self._evaluate_for = evaluate_for
        self._report_each = report_each
        self._report_verbose = os.getenv("VERBOSE", "1") not in ["", "0"]

        if all(hasattr(env.unwrapped, member) for member in ["ale", "seed_game", "load_game"]):  # Seed ALE.
            self.unwrapped.seed_game(seed)
            self.unwrapped.load_game()
        gym.Env.reset(self.unwrapped, seed=seed)
        self.action_space.seed(seed)
        self.observation_space.seed(seed)
        for passthrough in ["expert_trajectory", "expert_episode"]:
            if hasattr(env, passthrough):
                setattr(self, passthrough, getattr(env, passthrough))
            elif hasattr(env.unwrapped, passthrough):
                setattr(self, passthrough, getattr(env.unwrapped, passthrough))

        self._episode_running = False
        self._episode_returns = []
        self._evaluating_from = None
        self._original_render_mode = env.render_mode
        self._pygame = __import__("pygame") if self._render_each else None

    @property
    def episode(self):
        return len(self._episode_returns)

    def reset(self, *, start_evaluation=False, logging=True, seed=None, options=None):
        start_evaluation = start_evaluation or (options or {}).get("start_evaluation", False)
        logging = logging and (options or {}).get("logging", True)

        if seed is not None:
            raise RuntimeError("The EvaluationEnv cannot be reseeded")
        if self._evaluating_from is not None and self._episode_running:
            raise RuntimeError("Cannot reset a running episode after `start_evaluation=True`")
        if start_evaluation and self._evaluating_from is None:
            self._evaluating_from = self.episode

        if logging and self._render_each and (self.episode + 1) % self._render_each == 0:
            self.unwrapped.render_mode = "human"
        elif self._render_each:
            self.unwrapped.render_mode = self._original_render_mode
        self._episode_running = True
        self._episode_return = 0 if logging or self._evaluating_from is not None else None
        return super().reset(options=options)

    def step(self, action):
        if not self._episode_running:
            raise RuntimeError("Cannot run `step` on environments without an active episode, run `reset` first")

        observation, reward, terminated, truncated, info = super().step(action)
        done = terminated or truncated

        self._episode_running = not done
        if self._episode_return is not None:
            self._episode_return += reward
        if self._episode_return is not None and done:
            self._episode_returns.append(self._episode_return)

            if self._report_each and self.episode % self._report_each == 0:
                print("Episode {}, mean {}-episode return {:.2f} +-{:.2f}{}".format(
                    self.episode, self._evaluate_for, np.mean(self._episode_returns[-self._evaluate_for:]),
                    np.std(self._episode_returns[-self._evaluate_for:]), "" if not self._report_verbose else
                    ", returns " + " ".join(map("{:g}".format, self._episode_returns[-self._report_each:]))),
                    file=sys.stderr, flush=True)
            if self._evaluating_from is not None and self.episode >= self._evaluating_from + self._evaluate_for:
                print("The mean {}-episode return after evaluation {:.2f} +-{:.2f}".format(
                    self._evaluate_for, np.mean(self._episode_returns[-self._evaluate_for:]),
                    np.std(self._episode_returns[-self._evaluate_for:])), flush=True)
                self.close()
                sys.exit(0)

        if self._pygame and self.unwrapped.render_mode == "human" and self._pygame.get_init():
            if self._pygame.event.get(self._pygame.QUIT):
                self.unwrapped.render_mode = self._original_render_mode

        return observation, reward, terminated, truncated, info
